/*
	Fractal - A program to calculate Newton fractals.
	Copyright (C) 2004 Johannes Bauer
	
	This file is part of Fractal.

    Fractal is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Fractal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Fractal; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include <iostream>
#include <vector>
#include <complex.h>
#include <stdio.h>

#include "SDL.h"

#include "Fractal.hpp"
#include "CPoly.hpp"
#include "CSolutions.hpp"
#include "CNewton.hpp"

const double MaxError = 0.01;
const int MaxIterations = 1500;

CNewton::CNewton(CPoly Polynomial) {
	this->Polynomial=Polynomial;
	this->Polynomial_Derivation=Polynomial;
	this->Polynomial_Derivation.Differentiate();
	this->Calls=0;
	this->Iterations_Total=0;
	this->Iterations_Break=0;
	this->Iterations_Max=0;

	std::cout << "Function initialized:\n";
	this->Polynomial.Display();
	this->Polynomial.Display_Raw();
	std::cout << "\n";
	std::cout << "Derivation:\n";
	this->Polynomial_Derivation.Display();
	std::cout << "\n";
};

void CNewton::Reset() {
//	this->Solutions.Clear();
	this->Calls=0;
	this->Iterations_Total=0;
	this->Iterations_Break=0;
	this->Iterations_Max=0;
}

/*
Uint32 Colors[] = {
	0x140960,
	0x1ee844,
	0x1ecee8,
	0xdce81e,
	0xe20515,
	0xff8a00
};*/

Uint32 Colors[] = {
	0xff2121,
	0xff6b21,
	0xffb521,
	0xffff21,
	0xb5ff21,
	0x6bff21,
	0x21ff21,
	0x21ff6b,
	0x21ffb5,
	0x21ffff,
	0x21b5ff,
	0x216bff,
	0x2121ff,
	0x6b21ff,
	0xb521ff,
	0xff21ff,
	0xff21b5,
	0xff216b
};

Uint32 Map_Color(int Color) {
	Uint8 Red;
	Uint8 Green;
	Uint8 Blue;
	Uint32 Color_Table;

	Color%=(sizeof(Colors)/sizeof(Uint32));
	
	Color_Table=Colors[Color];
	Red=(Color_Table>>16)&(0xff);
	Green=(Color_Table>>8)&(0xff);
	Blue=(Color_Table>>0)&(0xff);
	
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	return ((Red<<24)&0xff000000) | ((Green<<16)&0x00ff0000) | ((Blue<<8)&0x0000ff00) | (0xff&0x000000ff);
#else
	return ((0xff<<24)&0xff000000) | ((Blue<<16)&0x00ff0000) | ((Green<<8)&0x0000ff00) | (Red&0x000000ff);
#endif
}


double absolute(double x) {
	if (x<0) return -x;
	return x;
}

cmplx *CNewton::Calculate(const cmplx *Start) {
	cmplx *Value;
	cmplx *Wert, *WertDer;
	int Iterations;
	double OldR, OldI, NewR, NewI;
	Value=new cmplx(*Start);

	NewR=Value->real();
	NewI=Value->imag();
	OldR=NewR+10;
	OldI=NewI+10;
	Iterations=0;
	while (1) {
		Wert=this->Polynomial.Value(Value);
		WertDer=this->Polynomial_Derivation.Value(Value);
		(*Value)-=((*Wert)/(*WertDer));
		delete Wert;
		delete WertDer;

		OldR=NewR;
		OldI=NewI;
		NewR=Value->real();
		NewI=Value->imag();
		Iterations++;
		if (Iterations>MaxIterations) {
			this->Iterations_Break++;
			break;
		}
		if ((absolute(OldR-NewR)<MaxError) && (absolute(OldI-NewI)<MaxError)) break;
	}
	this->Calls++;
	this->Iterations_Total+=Iterations;
	if ((unsigned long)Iterations>this->Iterations_Max) {
		this->Iterations_Max=Iterations;
	}
	return Value;
}

Uint32 CNewton::GetColor(const cmplx *Start) {
	int SolutionNumber;
	cmplx *Result;
	Result=this->Calculate(Start);
	SolutionNumber=this->Solutions.Get_Solution(*Result);
	delete Result;
	return Map_Color(SolutionNumber);
}

void CNewton::Analyze(const cmplx *Start) {
	cmplx *Result;
	Result=this->Calculate(Start);
	this->Solutions.Add_Solution(*Result);
	delete Result;
}

void CNewton::Display_Stats() {
	printf("Number of calculations: %lu\n",this->Calls);
	printf("Avg. iterations/call  : %.2f\n",(double)this->Iterations_Total/(double)this->Calls);
	printf("Maximum iterations    : %lu\n",this->Iterations_Max);
	printf("Unsharp calculations  : %lu\n",this->Iterations_Break);
}

void CNewton::Display_Solutions() {
	this->Solutions.Show();
}

void CNewton::Sort_Solutions() {
	this->Solutions.Sort();
}

CNewton::~CNewton() {
}

