/*
	Fractal - A program to calculate Newton fractals.
	Copyright (C) 2004 Johannes Bauer
	
	This file is part of Fractal.

    Fractal is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Fractal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Fractal; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include <cmath>
#include <complex.h>
#include <sstream>
#include <iostream>
#include <stdio.h>
#include <vector>
#include <stdlib.h>
#include <sys/timeb.h>
#include <string.h>
#include <errno.h>

#include "SDL.h"

#include "Fractal.hpp"
#include "CPoly.hpp"
#include "CSolutions.hpp"
#include "CNewton.hpp"
#include "CTimer.hpp"
#include "CTexExport.hpp"
#include "PngSave.h"

#define REGULAR_PUTPIXEL

const int REDRAW_INTERVAL = 100;
const double ZoomFactor = 1.5;
const double ZoomFactorHuge = 10.0;

bool Cluster_Active = false;
int Cluster_Number = 0;
int Cluster_Total = 0;
int Cluster_Count = 0;

#ifdef REGULAR_PUTPIXEL
void putpixel(SDL_Surface *surface, int x, int y, Uint32 pixel) {
  int bpp = surface->format->BytesPerPixel;
  /* Here p is the address to the pixel we want to set */
  Uint8 *p = (Uint8 *) surface->pixels + y * surface->pitch + x * bpp;

  switch (bpp) {
  case 1:
	*p = pixel;
	break;

  case 2:
	*(Uint16 *) p = pixel;
	break;

  case 3:
	if (SDL_BYTEORDER == SDL_BIG_ENDIAN) {
	  p[0] = (pixel >> 16) & 0xff;
	  p[1] = (pixel >> 8) & 0xff;
	  p[2] = pixel & 0xff;
	}
	else {
	  p[0] = pixel & 0xff;
	  p[1] = (pixel >> 8) & 0xff;
	  p[2] = (pixel >> 16) & 0xff;
	}
	break;

  case 4:
	*(Uint32 *) p = pixel;
	break;
  }
}
#else
void putpixel(SDL_Surface *surface, int x, int y, Uint32 pixel) {
	Uint8 *p = (Uint8 *) surface->pixels + y * surface->pitch + x * 4;
	*(Uint32 *) p = pixel;
}
#endif

FILE *ReadConf(char *Filename, int *ResolutionX, int *ResolutionY, CPoly *Term) {
	char Buffer[1024];
	FILE *f;
	char *Token;
	const char Delimiter[] = "\t ";
	int Cnt;
	
	f=fopen(Filename,"r");
	if (f==NULL) {
		return NULL;
	}

	fgets(Buffer,1023,f);
	Cnt=0;
	Token=strtok(Buffer,Delimiter);
	while (Token!=NULL) {
		if (Cnt==0) {
			*ResolutionX=atoi(Token);
		}
		if (Cnt==1) {
			*ResolutionY=atoi(Token);
		}
		Cnt++;
		Token=strtok(NULL,Delimiter);
	}

	fgets(Buffer,1023,f);
	Token=strtok(Buffer,Delimiter);
	while (Token!=NULL) {
		Term->Add_Coefficient(new cmplx(atof(Token)));
		Token=strtok(NULL,Delimiter);
	}

	return f;
}

bool ReadConf_Next(FILE *File, double *R1, double *I1, double *R2, double *I2, char *Name) {
	char Buffer[1024];
	char *Token;
	const char Delimiter[] = "\t ";
	int Cnt;
	
	fgets(Buffer,1023,File);
	Buffer[strlen(Buffer)-1]=0;
	Cnt=0;
	Token=strtok(Buffer,Delimiter);
	while (Token!=NULL) {
		if (Cnt==0) {
			*R1=atof(Token);
		}
		if (Cnt==1) {
			*I1=atof(Token);
		}
		if (Cnt==2) {
			*R2=atof(Token);
		}
		if (Cnt==3) {
			*I2=atof(Token);
		}
		if (Cnt==4) {
			strncpy(Name,Token,255);
			return true;
		}
		Cnt++;
		Token=strtok(NULL,Delimiter);
	}
	return false;
}

/*
void Fill(SDL_Surface *Surface, int X1, int Y1, int X2, int Y2, Uint32 Color) {
	int X, Y;
	for (X=X1;X<=X2;X++) {
		for (Y=Y1;Y<=Y2;Y++) {
			putpixel(Surface,X,Y,Color);
		}
	}
}

void Redraw_Box(SDL_Surface *Picture, SDL_Surface *MainScreen, int X1, int Y1, int X2, int Y2, double *MinR, double *MinI, double *MaxR, double *MaxI, CNewton *Newton) {
	int X, Y;
	int ResolutionX, ResolutionY;
	double Real, Imaginary;
	double StepR, StepI;
	Uint32 Color, Color_Reference=0;
	int Mode=0;
	bool First;
	bool All_Equal;
	cmplx *Wert;
	int MidX, MidY;

	if ((X1>X2) || (Y1>Y2)) return;
	
	ResolutionX=Picture->w;
	ResolutionY=Picture->h;
	StepR=(*MaxR-*MinR)/(double)ResolutionX;
	StepI=(*MaxI-*MinI)/(double)ResolutionY;
	
//	printf("Redrawing box %d,%d  %d,%d\n",X1,Y1,X2,Y2);
	
	X=X1;
	Y=Y1;
	First=true;
	All_Equal=true;
	while (Mode!=4) {
		// Get the pixel coordinates
		Real=*MinR+(X*StepR);
		Imaginary=*MinI+(Y*StepI);

		// Calculate the color of the pixel and draw it in
		Wert=new cmplx(Real,Imaginary);
		Color=Newton->GetColor(Wert);
		putpixel(Picture,X,(ResolutionY-Y)-1,Color);
		delete Wert;

		if (First) {
			Color_Reference=Color;
			First=false;
		} else {
			All_Equal=All_Equal && (Color==Color_Reference);
		}
		
		switch (Mode) {
			case 0: {
				X++;
				if (X>X2) {
					X--;
					Y++;
					Mode++;
				}
				break;
			}
			case 1: {
				Y++;
				if (Y>Y2) {
					Y--;
					X--;
					Mode++;
				}
				break;
			}
			case 2: {
				X--;
				if (X<X1) {
					X++;
					Y--;
					Mode++;
				}
				break;
			}
			case 3: {
				Y--;
				if (Y<Y1) {
					Mode++;
				}
				break;
			}
			default: break;
		}
	}

	if (All_Equal) {
		Fill(Picture,X1+1,Y1+1,X2-1,Y2-1,Color_Reference);
	} else {
		MidX=(X2-X1)/2;
		MidY=(Y2-Y1)/2;

		if ((X2-X1)*(Y2-Y1)>1024) {
			Redraw_Box(Picture,MainScreen,X1+1,Y1+1,MidX,MidY,MinR,MinI,MaxR,MaxI,Newton);
			Redraw_Box(Picture,MainScreen,X1+1,MidY+1,MidX,Y2-1,MinR,MinI,MaxR,MaxI,Newton);
			Redraw_Box(Picture,MainScreen,MidX+1,Y1+1,X2-1,MidY,MinR,MinI,MaxR,MaxI,Newton);
			Redraw_Box(Picture,MainScreen,MidX+1,MidY+1,X2-1,Y2-1,MinR,MinI,MaxR,MaxI,Newton);
		}
	}
	
	if ((X2-X1)*(Y2-Y1)>4096) {
		SDL_BlitSurface(Picture,NULL,MainScreen,NULL);
		SDL_Flip(MainScreen);
	}
}

void Redraw_Advanced(SDL_Surface *Picture, SDL_Surface *MainScreen, bool Interactive, CNewton *Newton, double *MinR, double *MinI, double *MaxR, double *MaxI) {
	// Select the first box
	Redraw_Box(Picture,MainScreen,0,0,Picture->w-1,Picture->h-1,MinR,MinI,MaxR,MaxI,Newton);

	SDL_BlitSurface(Picture,NULL,MainScreen,NULL);
	SDL_Flip(MainScreen);
}
*/

//Redraw(Picture,MainScreen,Interactive,Newton,MinR,MinI,MaxR,MaxI);
bool Redraw_Classic(SDL_Surface *Picture, SDL_Surface *MainScreen, bool Interactive, CNewton *Newton, double *MinR, double *MinI, double *MaxR, double *MaxI) {
	int X, Y;
	int ResolutionX, ResolutionY;
	double Real, Imaginary;
	double StepR, StepI;
	cmplx *Wert;
	bool Redraw;
	Uint32 SDL_Color, SDL_ColorReference=0;
	bool AllEqual, First;
	SDL_Event Event;

	ResolutionX=Picture->w;
	ResolutionY=Picture->h;

	StepR=((*MaxR)-(*MinR))/(double)ResolutionX;
	StepI=((*MaxI)-(*MinI))/(double)ResolutionY;
	
	First=true;
	AllEqual=true;
	Redraw=true;
	Real=*MinR;
	X=0;
	while ((X<ResolutionX) && (Redraw)) {
		Imaginary=*MinI;
		Y=0;
		while (Y<ResolutionY) {
			// Execute Newton
			Wert=new cmplx(Real,Imaginary);
			SDL_Color=Newton->GetColor(Wert);
			if (First) {
				SDL_ColorReference=SDL_Color;
				First=false;
			} else {
				AllEqual=AllEqual && (SDL_ColorReference==SDL_Color);
			}
			putpixel(Picture,X,(ResolutionY-Y)-1,SDL_Color);
			delete Wert;
			
			Imaginary+=StepI;
			Y++;
		}
		if ((Interactive) && (X%REDRAW_INTERVAL==0)) {
			SDL_BlitSurface(Picture,NULL,MainScreen,NULL);
			SDL_Flip(MainScreen);
		}
		Real+=StepR;
		X++;
		if (Interactive) {
			while (SDL_PollEvent(&Event)) {
		   	 	if ((Event.type==SDL_KEYDOWN) && (Event.key.keysym.sym==SDLK_ESCAPE)) {
					Redraw=false;
				}
			}
		}
	}
	if (Interactive) {
		SDL_BlitSurface(Picture,NULL,MainScreen,NULL);
		SDL_Flip(MainScreen);
	}

	return AllEqual;
}

void Display_Parameters(char **argv) {
	fprintf(stderr,"Wrong syntax. Please use:\n");
	fprintf(stderr,"%s random_imag [Resolution] [Degree]\n",argv[0]);
	fprintf(stderr,"%s random_real [Resolution] [Degree]\n",argv[0]);
	fprintf(stderr,"%s term [Resolution] [x^0] [x^1] [x^2] ...\n",argv[0]);
	fprintf(stderr,"%s term_inv [Resolution] [x^n] [x^n-1] ...\n",argv[0]);
	fprintf(stderr,"%s term_pos [Resolution] [x^n] [x^n-1] ... [RMin] [IMin] [RMax] [IMax]\n",argv[0]);
	fprintf(stderr,"%s file [config file]\n",argv[0]);
	fprintf(stderr,"\n");
	fprintf(stderr,"Where [Resolution] is of the form: [F][ResolutionX]x[ResolutionY]\n");
	fprintf(stderr,"A preceding 'F' signals to use full-screen mode.\n");
	fprintf(stderr,"\n");
	fprintf(stderr,"Examples:\n");
	fprintf(stderr,"%s random_imag 800x600 2              // ax^2 + bx + c           a,b,c in C\n",argv[0]);
	fprintf(stderr,"%s random_real f640x480 3             // ax^3 + bx^2 + cx + d    a,b,c,d in R\n",argv[0]);
	fprintf(stderr,"%s term f1280x1024 -2 0 0 3           // -2x^3 + 3\n",argv[0]);
	fprintf(stderr,"%s term 300x200 1 2 0 4               // x^3 + 2x^2 + 4\n",argv[0]);
	fprintf(stderr,"%s term_inv 300x200 1 2 0 4           // 4x^3 + 2x + 1\n",argv[0]);
	fprintf(stderr,"%s term_pos 300x200 1 2 0 4 -1 -1 1 1 // x^3 + 2x^2 + 4 from (-1,-1) to (1,1)\n",argv[0]);
	fprintf(stderr,"%s file 150x100 myfract.txt\n",argv[0]);
}

void Parse_Resolution(char *Cmd, int *ResolutionX, int *ResolutionY, Uint32 *Options) {
	char *SubString;
	char Delimiter[]="xX";
	
	if (strlen(Cmd)==0) {
		fprintf(stderr,"No resolution parameter given. Using defaults.\n");
		return;
	}
	if ((Cmd[0]=='F') || (Cmd[0]=='f')) {
		*Options|=SDL_FULLSCREEN;
		Cmd++;
	}
	SubString=strtok(Cmd,Delimiter);
	if (SubString!=NULL) {
		*ResolutionX=atoi(SubString);
		SubString=strtok(NULL,Delimiter);
		if (SubString!=NULL) {
			*ResolutionY=atoi(SubString);
		} else {
			fprintf(stderr,"No second resolution parameter given. Using defaults.\n");
		}
	} else {
		fprintf(stderr,"No resolution parameter given. Using defaults.\n");
		return;
	}
}

double rnd() {
	FILE *f;
	double n;
	unsigned long Number, Number_Max;
	Number_Max=(unsigned long)-1;
	f=fopen("/dev/urandom","r");
	if (f==NULL) {
		printf("Could not open random number device '/dev/urandom': %s\n",strerror(errno));
		return 1.0;
	}
	fread(&Number,1,sizeof(unsigned long),f);
	n=(((double)Number/(double)Number_Max)*6.0)-3.0;
	
	fclose(f);
	return n;
}

int FileExists(char *Filename) {
	FILE *f;
	f=fopen(Filename,"r");
	if (f==NULL) {
		return 0;
	} else {
		fclose(f);
		return 1;
	}
}

int main(int argc, char **argv) {
	CTimer *Timer, *Timer_Total;
	int Drawings=0;
	CPoly *Term;
	CNewton *Newton;
	cmplx *Wert;
	double MinR=-1.1, MaxR=1.1, StepR, Real;
	double MinI=-1.1, MaxI=1.1, StepI, Imaginary;
	double NewMin, NewMax;
	SDL_Surface *MainScreen=NULL;
	SDL_Surface *Picture;
	char Buffer[256];
	Uint32 rmask, gmask, bmask, amask;
	double RandomR, RandomI;
	bool Interactive;
	FILE *BatchFile=NULL;
	char Filename[256];
	bool AllEqual;
	FILE *DefinitionFile;
	int ResolutionX, ResolutionY;
	Uint32 Options;

	if (argc<3) {
		/* No parameters given (or just one) */
		fprintf(stderr,"Fractal requires at least 2 parameters.\n\n");
		Display_Parameters(argv);
		return 1;
	}

	Interactive=true;
	Term=new CPoly();
	if (strcmp(argv[1],"random_imag")==0) {
		time_t Time;
		int Degree;
		if (argc<4) {
			fprintf(stderr,"Fractal requires at least 3 parameters in 'random_imag' mode.\n\n");
			Display_Parameters(argv);
			return 1;
		}
		Degree=atoi(argv[3])+1;
		time(&Time);
		srand(Time);
		for (int i=1;i<=Degree;i++) {
			RandomR=(((double)rand()/(double)RAND_MAX)*6)-3;
			RandomI=(((double)rand()/(double)RAND_MAX)*6)-3;
			Term->Add_Coefficient(new cmplx(RandomR,RandomI));
		}
	} else {
		if (strcmp(argv[1],"random_real")==0) {
			time_t Time;
			int Degree;
			if (argc<4) {
				fprintf(stderr,"Fractal requires at least 3 parameters in 'random_real' mode.\n\n");
				Display_Parameters(argv);
				return 1;
			}
			
			Degree=atoi(argv[3])+1;
			time(&Time);
			srand(Time);
			for (int i=1;i<=Degree;i++) {
				RandomR=(((double)rand()/(double)RAND_MAX)*6)-3;
				Term->Add_Coefficient(new cmplx(RandomR));
			}
		} else {
			if (strcmp(argv[1],"term")==0) {
				if (argc<6) {
					fprintf(stderr,"Fractal requires at least 5 parameters in 'term' mode.\n\n");
					Display_Parameters(argv);
					return 1;
				}
				for (int i=argc-1;i>=3;i--) {
					Term->Add_Coefficient(new cmplx(atof(argv[i])));
				}
			} else {
				if (strcmp(argv[1],"file")==0) {
					Interactive=false;
					BatchFile=ReadConf(argv[2],&ResolutionX,&ResolutionY,Term);
					if (BatchFile==NULL) {
						printf("Could not read from file '%s': %s\n",argv[2],strerror(errno));
						return 3;
					};
					if (argc==5) {
						Cluster_Active=true;
						Cluster_Number=atoi(argv[3])-1;
						Cluster_Total=atoi(argv[4]);
						printf("Activating cluster mode, this is machine %d of %d.\n",Cluster_Number+1,Cluster_Total);
					}
				} else {
					if (strcmp(argv[1],"term_inv")==0) {
						if (argc<6) {
							fprintf(stderr,"Fractal requires at least 5 parameters in 'term_inv' mode.\n\n");
							Display_Parameters(argv);
							return 1;
						}
						for (int i=3;i<argc;i++) {
							Term->Add_Coefficient(new cmplx(atof(argv[i])));
						}
					} else {
						if (strcmp(argv[1],"term_pos")==0) {
							if (argc<10) {
								fprintf(stderr,"Fractal requires at least 9 parameters in 'term_pos' mode.\n\n");
								Display_Parameters(argv);
								return 1;
							}
							for (int i=3;i<argc-4;i++) {
								Term->Add_Coefficient(new cmplx(atof(argv[i])));
							}
							MinR=atof(argv[argc-4]);
							MinI=atof(argv[argc-3]);
							MaxR=atof(argv[argc-2]);
							MaxI=atof(argv[argc-1]);
						} else {
							Display_Parameters(argv);
							return 2;
						}
					}
				}
			}
		}
	}

	if (strcmp(argv[1],"file")!=0) {
		ResolutionX=640;
		ResolutionY=480;
		Options=0;
		Parse_Resolution(argv[2],&ResolutionX,&ResolutionY,&Options);
	}
	
	Newton=new CNewton(*Term);
	
	// The term is now known, now analyze it
	printf("Analyzing function...\n");
	for (Real=-1;Real<=1;Real+=0.1) {
		for (Imaginary=-1;Imaginary<=1;Imaginary+=0.1) {
			Wert=new cmplx(Real,Imaginary);
			Newton->Analyze(Wert);
		}
	}
	printf("Done.\n");
	Newton->Sort_Solutions();
	Newton->Display_Solutions();
	
	if (Interactive) {
		if (SDL_Init(SDL_INIT_VIDEO)<0) {
			fprintf(stderr,"SDL init failed: %s\n",SDL_GetError());
			return 1;
		};
		atexit(SDL_Quit);

		MainScreen=SDL_SetVideoMode(ResolutionX,ResolutionY,32,SDL_SWSURFACE | Options);
		if (MainScreen==NULL) {
			fprintf(stderr,"Couldn't set %dx%d video mode: %s\n",ResolutionX,ResolutionY,SDL_GetError());
			return 1;
		}
	}
	
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    rmask = 0xff000000;
    gmask = 0x00ff0000;
    bmask = 0x0000ff00;
    amask = 0x000000ff;
#else
    rmask = 0x000000ff;
    gmask = 0x0000ff00;
    bmask = 0x00ff0000;
    amask = 0xff000000;
#endif
	Picture=SDL_CreateRGBSurface(SDL_SWSURFACE,ResolutionX,ResolutionY,32,rmask,gmask,bmask,amask);
	if (Picture==NULL) {
		printf("Picture memory could not be allocated: %s\n",SDL_GetError());
		return 1;
	}

	if (Interactive) {
//		SDL_ShowCursor(SDL_DISABLE);
		sprintf(Buffer,"Fractals");
		SDL_WM_SetCaption(Buffer,NULL);
		SDL_EnableUNICODE(1);
	}
	
	bool Quit=false;
	bool Redraw=true;
	SDL_Event Event;
	Timer=new CTimer();
	Timer_Total=new CTimer();
	while (!Quit) {
		if (!Interactive) {
			// Batch processing
			if (ReadConf_Next(BatchFile,&MinR,&MinI,&MaxR,&MaxI,Filename)) {
				if (Cluster_Active) {
					Cluster_Count++;
					if ((Cluster_Count-1)%Cluster_Total!=Cluster_Number) continue;
				}
				
				printf("Now calculating from (%.8f,%.8f) to (%.8f,%.8f), saving in '%s'...\n",MinR,MinI,MaxR,MaxI,Filename);
				fflush(stdout);		// So the output is generated
				fflush(stderr);
				if (FileExists(Filename)) {
					printf("File '%s' exists already, skipping...\n\n",Filename);
					continue;
				}
				Redraw=true;
			} else {
				printf("Batch processing ended.\n");
				break;
			}
		}
		
		if (Redraw) {
			printf("Drawing inaccuracy     : %.46f\n",MaxR-MinR < MaxI-MinI ? MaxR-MinR : MaxI-MinI);
			Drawings++;
			Timer->Reset();
			Newton->Reset();
			AllEqual=Redraw_Classic(Picture,MainScreen,Interactive,Newton,&MinR,&MinI,&MaxR,&MaxI);
			if (Interactive) {
				// Write information to file
				DefinitionFile=fopen("Information.txt","w");
				if (DefinitionFile==NULL) {
					printf("Could not open Information.txt for writing: %s\n",strerror(errno));
				} else {
					Term->Write(DefinitionFile);
					fprintf(DefinitionFile,"%.46f %.46f %.46f %.46f\n",MinR,MinI,MaxR,MaxI);
					fclose(DefinitionFile);
				}
			}
			if ((!Interactive) && (AllEqual)) {
				fprintf(stderr,"Canceling script procession due to all equal pixels!\n");
				Quit=true;
			}
		//	Redraw_Advanced(Picture,MainScreen,Interactive,Newton,&MinR,&MinI,&MaxR,&MaxI);
			Newton->Display_Stats();
			Redraw=false;
			if (Timer->Get()>1.0) {
				printf("Calculation time      : %.3f seconds.\n\n",Timer->Get());
			} else {
				printf("Framerate             : %.1f fps\n\n",1.0/Timer->Get());
			}
		}
		if (Interactive) {
			SDL_SavePNG(Picture,"Out.png");
			CTexExport *Tex;
			Tex=new CTexExport();
			Tex->Set_Coordinates(MinR,MinI,MaxR,MaxI);
			Tex->Set_Polynomial(Term);
			Tex->Write("Out.tex");
			delete(Tex);
		} else {
			SDL_SavePNG(Picture,Filename);
		}
	   
		while ((!Quit) && (!Redraw) && (Interactive)) {
			while (SDL_PollEvent(&Event)) {
				if (Event.type==SDL_MOUSEBUTTONDOWN) {
					if ((Event.button.button==1) ||
						(Event.button.button==4) ||
						(Event.button.button==5)) {
						StepR=(MaxR-MinR);										// Width
						Real=((double)Event.button.x/(double)ResolutionX);		// Click position [0..1]
						NewMin=(Real*StepR)+MinR-(StepR/2.0);
						NewMax=(Real*StepR)+MinR+(StepR/2.0);
						MinR=NewMin;
						MaxR=NewMax;

						StepI=(MaxI-MinI);											// Width
						Imaginary=1-((double)Event.button.y/(double)ResolutionY);	// Click position [0..1]
						NewMin=(Imaginary*StepI)+MinI-(StepI/2.0);
						NewMax=(Imaginary*StepI)+MinI+(StepI/2.0);
						MinI=NewMin;
						MaxI=NewMax;
						
						Redraw=true;
					}
					if (Event.button.button==4) {
						// Mousewheel up, scroll & zoom in
						Real=MaxR-MinR;			// Range
						StepR=(MinR+MaxR)/2.0;	// Midpoint
						Real/=ZoomFactor;
						MinR=StepR-(Real/2.0);
						MaxR=StepR+(Real/2.0);
							
						Imaginary=MaxI-MinI;	// Range
						StepI=(MinI+MaxI)/2.0;	// Midpoint
						Imaginary/=ZoomFactor;
						MinI=StepI-(Imaginary/2.0);
						MaxI=StepI+(Imaginary/2.0);
					}
					if (Event.button.button==5) {
						// Mousewheel down, scroll & zoom out
						Real=MaxR-MinR;			// Range
						StepR=(MinR+MaxR)/2.0;	// Midpoint
						Real*=ZoomFactor;
						MinR=StepR-(Real/2.0);
						MaxR=StepR+(Real/2.0);
						
						Imaginary=MaxI-MinI;	// Range
						StepI=(MinI+MaxI)/2.0;	// Midpoint
						Imaginary*=ZoomFactor;
						MinI=StepI-(Imaginary/2.0);
						MaxI=StepI+(Imaginary/2.0);
					}
					if (Event.button.button==2) {
						// Mousewheel press, restore defaults
						MinR=-1.1;
						MinI=-1.1;
						MaxR=1.1;
						MaxI=1.1;
						Redraw=true;
					}
					if (Redraw) printf("Mouse movement, new range: %.46f %.46f %.46f %.46f\n",MinR,MinI,MaxR,MaxI);
				}
		   	 	if (Event.type==SDL_KEYDOWN) {
					switch (Event.key.keysym.sym) {
		            	case SDLK_ESCAPE: {
							Quit=true;
							break;
						 }
		            	case SDLK_RETURN: {
							Redraw=true;
							break;
						 }
						case SDLK_DOWN: {
							Redraw=true;
							MinI+=0.2;
							MaxI+=0.2;
							printf("Keyboard movement, new range: %.46f %.46f %.46f %.46f\n",MinR,MinI,MaxR,MaxI);
							break;
						}
						case SDLK_UP: {
							Redraw=true;
							MinI-=0.2;
							MaxI-=0.2;
							printf("Keyboard movement, new range: %.46f %.46f %.46f %.46f\n",MinR,MinI,MaxR,MaxI);
							break;
						}
						case SDLK_LEFT: {
							Redraw=true;
							MinR-=0.2;
							MaxR-=0.2;
							printf("Keyboard movement, new range: %.46f %.46f %.46f %.46f\n",MinR,MinI,MaxR,MaxI);
							break;
						}
						case SDLK_RIGHT: {
							Redraw=true;
							MinR+=0.2;
							MaxR+=0.2;
							printf("Keyboard movement, new range: %.46f %.46f %.46f %.46f\n",MinR,MinI,MaxR,MaxI);
							break;
						}
						case SDLK_PLUS: {
							Real=MaxR-MinR;			// Range
							StepR=(MinR+MaxR)/2.0;	// Midpoint
							Real/=ZoomFactor;
							MinR=StepR-(Real/2.0);
							MaxR=StepR+(Real/2.0);
							
							Imaginary=MaxI-MinI;	// Range
							StepI=(MinI+MaxI)/2.0;	// Midpoint
							Imaginary/=ZoomFactor;
							MinI=StepI-(Imaginary/2.0);
							MaxI=StepI+(Imaginary/2.0);
						
							printf("Zooming in, new range: %.46f %.46f %.46f %.46f\n",MinR,MinI,MaxR,MaxI);
							Redraw=true;
							break;
						}
						case SDLK_MINUS: {
							Real=MaxR-MinR;			// Range
							StepR=(MinR+MaxR)/2.0;	// Midpoint
							Real*=ZoomFactor;
							MinR=StepR-(Real/2.0);
							MaxR=StepR+(Real/2.0);
							
							Imaginary=MaxI-MinI;	// Range
							StepI=(MinI+MaxI)/2.0;	// Midpoint
							Imaginary*=ZoomFactor;
							MinI=StepI-(Imaginary/2.0);
							MaxI=StepI+(Imaginary/2.0);
							
							printf("Zooming out, new range: %.46f %.46f %.46f %.46f\n",MinR,MinI,MaxR,MaxI);
							Redraw=true;
							break;
						}
						case SDLK_PAGEUP: {
							Real=MaxR-MinR;			// Range
							StepR=(MinR+MaxR)/2.0;	// Midpoint
							Real/=ZoomFactorHuge;
							MinR=StepR-(Real/2.0);
							MaxR=StepR+(Real/2.0);
							
							Imaginary=MaxI-MinI;	// Range
							StepI=(MinI+MaxI)/2.0;	// Midpoint
							Imaginary/=ZoomFactorHuge;
							MinI=StepI-(Imaginary/2.0);
							MaxI=StepI+(Imaginary/2.0);
						
							printf("Zooming in, new range: %.46f %.46f %.46f %.46f\n",MinR,MinI,MaxR,MaxI);
							Redraw=true;
							break;
						}
						case SDLK_PAGEDOWN: {
							Real=MaxR-MinR;			// Range
							StepR=(MinR+MaxR)/2.0;	// Midpoint
							Real*=ZoomFactorHuge;
							MinR=StepR-(Real/2.0);
							MaxR=StepR+(Real/2.0);
							
							Imaginary=MaxI-MinI;	// Range
							StepI=(MinI+MaxI)/2.0;	// Midpoint
							Imaginary*=ZoomFactorHuge;
							MinI=StepI-(Imaginary/2.0);
							MaxI=StepI+(Imaginary/2.0);
							
							printf("Zooming out, new range: %.46f %.46f %.46f %.46f\n",MinR,MinI,MaxR,MaxI);
							Redraw=true;
							break;
						}
						default: break;
					}
				}
			}
			SDL_Delay(100);
		}
		fflush(stdout);		// So the output is generated
		fflush(stderr);
	}

	if (!Interactive) {
		printf("Pixels/second                     : %.0f\n",(double)(ResolutionX*ResolutionY*Drawings)/Timer_Total->Get());
		printf("Total time                        : %.3f sec\n",Timer_Total->Get());
		printf("Total time in hms                 : %d:%02d:%02d\n",(int)Timer_Total->Get()/60/60,(int)Timer_Total->Get()/60%60,(int)Timer_Total->Get()%60%60);
		printf("Total time/picture [%4d pictures]: %.3f sec\n",Drawings,Timer_Total->Get()/(double)Drawings);
	}

	delete Term;
	delete Timer_Total;
	delete Timer;
	if (!Interactive) SDL_Quit();
	return 0;
}
